'''
FlavourMod (v1.1.2)
~~~~~~~~~~~~~~~~~
PURPOSE: 
Have certain civs start in regions matching their 'flavour', but interfere with
the stuff the base mapscript is doing as minimally as necessary. 'Flavour' 
in this context means everything you may associate with a certain civ and its 
typical landscape. Let it be based on its historical background or something 
like 'Elves -> treehugging -> forest' or 'Egypt -> sand -> desert'. These 
choices can be very subjective, so feel free to modify them to your liking (see
SETTINGS block below, it's rather self-explainatory, I hope.

ADITIONAL NOTE:
FlavorMod does not change the map itself, it only uses the starting locations
already provided by the base mapscript and redistributes them. Hence, if there 
is no suitable place for a particular civ, it will be set to start somewhere, 
but probably in a rather ill-fitting area. Albeit there will be a clear tendency,
no guarantees can be made, that all starts will be 'flavoury' for everybody.
Another small drawback: Teams won't start near each other in many cases. Would 
be pretty complex to assure flavour and close team members at the same time. 
Take a look at Quadratic Assignment Optimization Problems and you'll agree.

HOW-TO MAKE IT WORK:
It is very simple to use this module in a map script of your choice. Just copy 
this file into your "<civ4>/PublicMaps" folder, then open the map script file 
(e.g. Continents.py) with a text editor and copypaste the small code block below 
this paragraph into the mapscript (preferably at the top), and uncomment the 
'import' line by removing the '#'. Of course you're advised to make a backup of 
the original mapscript first and rename your modified version to something like 
"Continents_Flavoured.py" so that you don't confuse the various versions. 

copy below; remove '#' in 'from...import' line:
###########################################################
## This mapscript is modified to use the "FlavourMod".      
## See "FlavourMod.py" in the same directory for details.  
###########################################################
#from FlavourMod import *                                 
###########################################################
copy above

Hope you'll enjoy the mod. Andr. (Jean Elcard)

CHANGELOG:

Verson 1.1.0

	* Add: flavour information for Dural and Sheaim 
	* Add: new flavour parameter: FoodFocus (high food yield of starting location)
	* Add: new flavour parameter: ProductionFocus (like food focus, but for hammers)
	* Add: new flavour parameter: CommerceFocus (like food focus, but for commerce)
	* Add: new flavour parameter: MobilityFocus (prefer flat land without features)
	* Add: allow different shaped search areas (fatcross or square)
	* Change: moved FlavourMod to a later stage of the map generation process
	* Change: allow some of the normalization functions for buffing dull starts
	* Change: fine tuning of the flavour info of all civs (see the current list) 
	* Fix: civ-specific search areas will take effect now
	* Other: many code improvements, some more docu 
	
Version 1.1.1

	* Fix: no python exceptions any more if flavour parameter "improvements" is set
	* Add: flavour Info for Bannor (incense)
	* Add: flavour Info for Clan of Embers (jungle, plains, desert, Pyre of the Seraphic)
	* Add: additional flavour info for Illians (Letum Frigus) 
	* Add: additional flavour info for Elohim (sacred sites except Brocken Sepulcher) 
	
Version 1.1.2

	* Fix: a starting loction near world's end (map boarder) could break FlavourMod
	* Add: flavour Info for Kahdi (same for the renegades as for the Amurites)

'''

###############################################################################

from CvPythonExtensions import *
import CvMapGeneratorUtil
import CvUtil
import PyHelpers
import sys

gc = CyGlobalContext()
PyPlayer = PyHelpers.PyPlayer

__version__   = "1.1.2"
__author__    = "Andr Gtze (Jean Elcard), JeanElcard@digital-filestore.de"

__all__ =  ["normalizeStartingPlotLocations",
			#"normalizeAddRiver",
			"normalizeRemovePeaks",
			#"normalizeAddLakes",
			"normalizeRemoveBadFeatures",
			"normalizeRemoveBadTerrain",
			#"normalizeAddFoodBonuses",
			"normalizeAddGoodTerrain",
			#"normalizeAddExtras",
			"startHumansOnSameTile"]

			
###############################################################################

#SEARCH AREA PARAMETERS: 

HINTERLAND = 6 	# Hinterland is the area around starting plots which 
				# will be checked for plots matching a civ's flavour.

HEARTLAND = 2 	# Heartland is a subset area of Hinterland. Matches inside 
				# heartland get full weight; Flavour matches outside heartland,
				# but still inside hinterland get a linear decreasing weight. 
				# This means, a match counts less and less the farther away and
				# zero outside of hinterland

	# The example below visualizes the meaning of the two values above:
				
		# Search Area shape for the example values 3, 2, 1 and 0:

		#		|3|3|3|
		#	  |3|2|2|2|3|		  |2|2|2|
		#	|3|2|1|1|1|2|3|		|2|1|1|1|2|		|1|1|1|
		#	|3|2|1|0|1|2|3|		|2|1|0|1|2|		|1|0|1|		|0|
		#	|3|2|1|1|1|2|3|		|2|1|1|1|2|		|1|1|1|
		#	  |3|2|2|2|3|		  |2|2|2|
		#		|3|3|3|
		
	# As you can see the default area shape is a fatcross. You might change 
	# its size for certain civs (see the Kuriotates) or even change the shape
	# from fatcross to square for some other civs (see the Hippus).

#hard coded Fall from Heaven Unique improvements list:
	
UNIQUE_IMPROVEMENTS = [	"IMPROVEMENT_REMNANTS_OF_PATRIA",
						"IMPROVEMENT_STANDING_STONES",
						"IMPROVEMENT_TOMB_OF_SUCELLUS",
						"IMPROVEMENT_YGGDRASIL",
						"IMPROVEMENT_BROKEN_SEPULCHER",
						"IMPROVEMENT_MIRROR_OF_HEAVEN",
						"IMPROVEMENT_DRAGON_BONES",
						"IMPROVEMENT_LETUM_FRIGUS",
						"IMPROVEMENT_MAELSTROM",
						"IMPROVEMENT_ODIOS_PRISON",
						"IMPROVEMENT_POOL_OF_TEARS",
						"IMPROVEMENT_PYRE_OF_THE_SERAPHIC"]	
	
#DEBUGGING PARAMETERS:
	
DEBUG = False	# If "True", debug information will be logged to PythonDbg.log.
				# Requires an enabled logging system too. (Civilization4.ini)

SPELLCHECK = False	# If "True", you will get python errors, if using 
					# non-existing civnames etc in the settings above. Don't 
					# activate, if you want to use FlavourMod with different
					# mods. Otherwise you'll get errors, because a civ from 
					# one mod doesn't exist in another and so on.

###############################################################################
					
def doFlavourModStuff(): # This function redistributes the starting locations. 
	
	NiTextOut("Python: Redistributing starting positions (FfH_script_version) ...")	
	dCivFlavourInfos = {}
	
	'''
	SETTINGS:
	
	Some remarks first:
	
	Flavour Parameters (for their default values see class CivFlavourInfo):	
	
		plottypes, terrains, features, bonuses, improvements
		bFoodFocus,	bProductionFocus, bCommerceFocus, bMobilityFocus
		fCoastalStartValue, pSearchArea
		
	List of Civilisations you might want to give some flavour:
		
		Fall from Heaven:
		
			AMURITES, BALSERAPHS, BANNOR, CALABIM, CLAN_OF_EMBERS, 
			DOVIELLO, ELOHIM, GRIGORI, HIPPUS, ILLIANS, 
			KHAZAD, KURIOTATES, LANUN, LJOSALFAR, LUCHUIRP, 
			MALAKIM, SHEAIM, SIDAR, SVARTALFAR
	
		Fall Further: 
		
			ARCHOS, CHISLEV, CUALLI, DURAL, KAHDI, MAZATL
		
	The correct names for terrains etc. you can get from the XML files.
	
	Make your flavour changes and additions in the sections below:
	'''
	
	## Fall From Heaven Settings:
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_AMURITES")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.bonuses[getInfoType("BONUS_MANA")] = 1.0
	pCivFlavourInfo.bonuses[getInfoType("BONUS_REAGENTS")] = 1.5
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_BANNOR")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.bonuses[getInfoType("BONUS_INCENSE")] = 2.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_CALABIM")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.bFoodFocus = True
	pCivFlavourInfo.pSearchArea = SearchArea(FatCrossArea(2), FatCrossArea(2))
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_CLAN_OF_EMBERS")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.improvements[getInfoType("IMPROVEMENT_PYRE_OF_THE_SERAPHIC")] = 100.0
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_DESERT")] = 1.0
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_PLAINS")] = 1.5
	pCivFlavourInfo.features[getInfoType("FEATURE_JUNGLE")] = 2.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_DOVIELLO")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_TUNDRA")] = 1.0
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_SNOW")] = 0.1
	pCivFlavourInfo.features[getInfoType("FEATURE_ICE")] = 0.1
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_ELOHIM")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.bonuses[getInfoType("BONUS_INCENSE")] = 1.0
	for sUniqueImprovement in UNIQUE_IMPROVEMENTS: 
		if not sUniqueImprovement == "IMPROVEMENT_BROKEN_SEPULCHER":
			pCivFlavourInfo.improvements[getInfoType(sUniqueImprovement)] = 5.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_HIPPUS")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.bMobilityFocus = True
	pCivFlavourInfo.pSearchArea = SearchArea(SquareArea(0), SquareArea(5))
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_ILLIANS")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.improvements[getInfoType("IMPROVEMENT_LETUM_FRIGUS")] = 100.0
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_TUNDRA")] = 1.0
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_SNOW")] = 0.1
	pCivFlavourInfo.features[getInfoType("FEATURE_ICE")] = 0.1
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_KHAZAD")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.plottypes[PlotTypes.PLOT_HILLS] = 1.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_KURIOTATES")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_COAST")] = -0.5
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_OCEAN")] = -1.0
	pCivFlavourInfo.pSearchArea = SearchArea(FatCrossArea(3), FatCrossArea(6))
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_LANUN")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.fCoastalStartValue = 15.0
	pCivFlavourInfo.bonuses[getInfoType("BONUS_PEARL")] = 5.0
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_COAST")] = 1.0
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_OCEAN")] = 0.5
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_LJOSALFAR")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.features[getInfoType("FEATURE_FOREST")] = 1.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_LUCHUIRP")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.plottypes[PlotTypes.PLOT_HILLS] = 0.5
	pCivFlavourInfo.bProductionFocus = True
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_MALAKIM")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_DESERT")] = 1.0
	pCivFlavourInfo.features[getInfoType("FEATURE_OASIS")] = 1.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_SHEAIM")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.bonuses[getInfoType("BONUS_MANA")] = 1.0
	pCivFlavourInfo.bonuses[getInfoType("BONUS_REAGENTS")] = 1.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_SVARTALFAR")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.features[getInfoType("FEATURE_FOREST")] = 1.0
	pCivFlavourInfo.bonuses[getInfoType("BONUS_MANA")] = 1.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
		
		
	## Fall Further Settings:
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_ARCHOS")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.bFoodFocus = True
	pCivFlavourInfo.pSearchArea = SearchArea(FatCrossArea(2), FatCrossArea(2))
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_CHISLEV")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_PLAINS")] = 1.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_CUALLI")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.features[getInfoType("FEATURE_JUNGLE")] = 1.5
	pCivFlavourInfo.features[getInfoType("FEATURE_FOREST")] = 0.5
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_GRASS")] = 0.5
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_TUNDRA")] = -1.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_DURAL")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.bonuses[getInfoType("BONUS_MARBLE")] = 1.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_KAHDI")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.bonuses[getInfoType("BONUS_MANA")] = 1.0
	pCivFlavourInfo.bonuses[getInfoType("BONUS_REAGENTS")] = 1.5
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	iCivilization = getInfoType("CIVILIZATION_MAZATL")
	pCivFlavourInfo = CivFlavourInfo(iCivilization)
	pCivFlavourInfo.features[getInfoType("FEATURE_JUNGLE")] = 1.5
	pCivFlavourInfo.features[getInfoType("FEATURE_FOREST")] = 0.5
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_GRASS")] = 0.5
	pCivFlavourInfo.terrains[getInfoType("TERRAIN_TUNDRA")] = -1.0
	dCivFlavourInfos[iCivilization] = pCivFlavourInfo
	###########################################################################
	
	pStartingPlots = []
	pFlavourCivPlayers = []
	pOrdinaryCivPlayers = []
	
	iPlayers = gc.getGame().countCivPlayersEverAlive()
	
	# create lists of flavoured and ordinary civ players in game
	for iPlayer in range(iPlayers):
		pPlayer = gc.getPlayer(iPlayer)
		iCivType = pPlayer.getCivilizationType()
		if iCivType in dCivFlavourInfos.keys():
			pFlavourCivPlayers.append(pPlayer)
		else:
			pOrdinaryCivPlayers.append(pPlayer)
	
	# return with no changes, if there is no flavoured civ in this game
	debugOut("Flavoured Civilizations in this game: %d" % (len(pFlavourCivPlayers)))
	if len(pFlavourCivPlayers) == 0: return
		
	# create a list of the starting plots currently assigned to each player
	for iPlayer in range(iPlayers):
		pPlayer = gc.getPlayer(iPlayer)
		pStartingPlot = pPlayer.getStartingPlot()	  
		pStartingPlots.append(pStartingPlot)
	
	# create a "mxn" profit matrix (Flavoured Civilizations x Starting Plots)
	profit_matrix = []
	for pPlayer in pFlavourCivPlayers:
		scores = []
		pCivFlavourInfo	= dCivFlavourInfos[pPlayer.getCivilizationType()]
		debugOut("Commence checking plots for %s:" % gc.getCivilizationInfo(pPlayer.getCivilizationType()).getDescription())
		for pStartingPlot in pStartingPlots:
			score = pCivFlavourInfo.evaluateStartingPlot(pStartingPlot, pPlayer)
			scores.append(score)
		profit_matrix.append(scores)
	debugOut(strMatrix(profit_matrix, '1: mxn Profit Matrix'))
		
	# normalize rows of profit matrix to make scores comparable between civs
	profit_matrix = normalizeMatrix(profit_matrix)
	debugOut(strMatrix(profit_matrix, '2: mxn Profit Matrix (n)'))
		
	# transform profit matrix to  square cost matrix to match requirements for munkres
	cost_matrix = squarizeMatrix(createCostMatrix(profit_matrix), 0.0)
	debugOut(strMatrix(cost_matrix, '3: nxn Cost Matrix'))
	
	# run Linear Assignment Algorithm (Hungarian Method by Munkres)  
	m = Munkres()
	optimizationResults = m.compute(cost_matrix)
	total = 0
	for i, j in optimizationResults:
	    value = cost_matrix[i][j]
	    total += value
	debugOut('total cost: %f' % total)
	
	# reallocate starting positions to civs according to optimization results:
	pPlayers = pFlavourCivPlayers + pOrdinaryCivPlayers
	for x, y in optimizationResults:
		pPlayers[x].setStartingPlot(pStartingPlots[y], True)
		sCivName = getCivilizationName(pPlayers[x].getCivilizationType())
		debugOut('(%d, %d) %s -> %f' % (x, y, sCivName, cost_matrix[x][y]))
	
	return False

	
###############################################################################	
## The functions in the section below are changing the terrain, features etc 
## in the vicinity of starting locations AFTER the assignment of the original 
## starting locations has taken place. I've disabled some of them, because 
## peaks, "bad terrain" can be very flavourful indeed. 

iCallCounter = 0	# Global varibale counting the external calls of 
					# startHumansOnSameTile(). Necessary, because we want 
					# to redistribiute the starting locations only once.

def normalizeStartingPlotLocations():
	debugOut("CallCounter Reset")
	global iCallCounter
	iCallCounter = 0 
	return None
	
def normalizeAddRiver(): # allowed
	return None
	
def normalizeRemovePeaks():
	return None
	
def normalizeAddLakes(): # allowed
	return None

def normalizeRemoveBadFeatures():
	return None

def normalizeRemoveBadTerrain():
	return None
	
def normalizeAddFoodBonuses(): # allowed
	return None

def normalizeAddGoodTerrain():
	return None

def normalizeAddExtras(): # allowed
	return None

# FlavourMod kicks in here 
def startHumansOnSameTile():
	global iCallCounter
	iCallCounter += 1
	debugOut("%d. Call of FlavourMod:" % iCallCounter)
	if iCallCounter == 1: 
		doFlavourModStuff()
	elif iCallCounter == 2: 
		fixFirstHumanUnitLocation() 
	return False
'''
Problem: 	First human unit gets placed at original starting location.
Workaround: Veil map and move unit to the new flavoured starting location.
			The function below takes care of this:	
'''			
def fixFirstHumanUnitLocation():
	map = CyMap() 
	player = PyPlayer(0)
	pPlayer = player.getPlayer()
	pPlot = pPlayer.getStartingPlot()
	for pUnit in player.getUnitList():
		eTeam = TeamTypes(pPlayer.getTeam())
		map.setRevealedPlots(eTeam, False, False)
		pUnit.setXY(pPlot.getX(), pPlot.getY(), False, True, True)	
	return None	
	
###############################################################################
## The Flavour Class and helper functions for handling matrices and stuff.

class CivFlavourInfo: 

	def __init__(self, civ):
		self.civilization = civ
		self.terrains = {}
		self.features = {}
		self.bonuses = {}
		self.improvements = {}
		self.plottypes = {}
		self.bFoodFocus = False
		self.bProductionFocus = False
		self.bCommerceFocus = False
		self.bMobilityFocus = False
		self.fCoastalStartValue = 0.0 
		self.pSearchArea = SearchArea(FatCrossArea(HEARTLAND), FatCrossArea(HINTERLAND))
		
	def evaluateStartingPlot(self, pStartingPlot, pPlayer):
	
		debugOut("Evaluate Starting Plot (%d, %d)." % (pStartingPlot.getX(), pStartingPlot.getY()))
	
		fStartingAreaScore = 0.0
		pMap = CyMap() 
				
		if self.fCoastalStartValue > 0:
			if pStartingPlot.isCoastalLand(): 
				fStartingAreaScore += self.fCoastalStartValue
		
		for x, y in self.pSearchArea.getPlots():
			
			iPlotScore = 0		
			pPlot = pMap.plot(pStartingPlot.getX() + x, pStartingPlot.getY() + y)
			if pPlot.isNone(): continue
			
			eTerrain = pPlot.getTerrainType()
			ePlot = pPlot.getPlotType()
			eFeature = pPlot.getFeatureType()
			eBonus = pPlot.getBonusType(pPlayer.getTeam())
			eImprovement = pPlot.getImprovementType()
					
			if eTerrain in self.terrains.keys():
				iPlotScore += self.terrains[eTerrain]
			if ePlot in self.plottypes.keys():
				iPlotScore += self.plottypes[ePlot]
			if eFeature in self.features.keys():
				iPlotScore += self.features[eFeature]
			if eBonus in self.bonuses.keys():
				iPlotScore += self.bonuses[eBonus]
			if eImprovement in self.improvements.keys():
				iPlotScore += self.improvements[eImprovement]
				debugOut("Found %s with weight %f." % (gc.getImprovementInfo(eImprovement).getDescription(), self.pSearchArea.getPlotWeight(x,y)))
				
			if self.bFoodFocus:
				iPlotScore += getPlotYield(pPlot, YieldTypes.YIELD_FOOD, TeamTypes(pPlayer.getTeam()))
			if self.bProductionFocus:
				iPlotScore += getPlotYield(pPlot, YieldTypes.YIELD_PRODUCTION, TeamTypes(pPlayer.getTeam()))
			if self.bCommerceFocus:				
				iPlotScore += getPlotYield(pPlot, YieldTypes.YIELD_COMMERCE, TeamTypes(pPlayer.getTeam()))
				
			if self.bMobilityFocus:
				iPlotScore -= getMovementCost(pPlot)				
				
			fStartingAreaScore += iPlotScore * self.pSearchArea.getPlotWeight(x,y)
		
		return fStartingAreaScore

		
def getPlotYield(pPlot, eYield, eTeam):
	iYield = 0
	iYieldChanges = []
	eBonus = pPlot.getBonusType(eTeam)
	iYield = pPlot.calculateNatureYield(eYield, eTeam, False) 
	if eBonus != BonusTypes.NO_BONUS:
		for iImprovement in range(gc.getNumImprovementInfos()):
			pImprovement = gc.getImprovementInfo(iImprovement)
			if pImprovement.isImprovementBonusMakesValid(eBonus):
				iYieldChange = pImprovement.getImprovementBonusYield(eBonus, eYield)
				iYieldChanges.append(iYieldChange)
	if len(iYieldChanges) > 0:
		iYield += max(iYieldChanges)
	return iYield

	
def getMovementCost(pPlot):
	iRegularCost = 0
	if pPlot.isWater() or pPlot.isImpassable():	
		return 2
	if pPlot.getFeatureType() != FeatureTypes.NO_FEATURE:
		iRegularCost = gc.getFeatureInfo(pPlot.getFeatureType()).getMovementCost()
	else: 
		iRegularCost = gc.getTerrainInfo(pPlot.getTerrainType()).getMovementCost()
	if pPlot.isHills():
		return iRegularCost + gc.getHILLS_EXTRA_MOVEMENT()
	return iRegularCost

	
def getInfoType(string):
	iInfoType = gc.getInfoTypeForString(string)
	if iInfoType == -1 and SPELLCHECK:
		arg = ("%s unknown! Probably just a Typing Error." % string)
		raise ValueError, arg
	return iInfoType
	
	
def getCivilizationName(iCivilizationInfoType):
	return gc.getCivilizationInfo(iCivilizationInfoType).getType()
	
	
def normalizeMatrix(matrix):
	norm_matrix = []
	dimX = len(matrix)
	dimY = len(matrix[0])
	for i in range(dimX):
		minScore = min(matrix[i])
		maxScore = max(matrix[i])
		norm_matrix.append([])
		if minScore != maxScore:
			for j in range(dimY):
				a = (matrix[i][j] - minScore)
				b = float(maxScore - minScore)
				norm_matrix[i].append(a/b)
		else:
			for j in range(dimY):
				norm_matrix[i].append(1)
	return norm_matrix
	
	
def createCostMatrix(matrix):
	cost_matrix = []
	dimX = len(matrix)
	dimY = len(matrix[0])
	maxElement = max(matrix[0])
	for row in matrix:
		if max(row) > maxElement:
			maxElement = max(row)
	for i in range(dimX):
		cost_matrix.append([])
		for j in range(dimY):
			cost_matrix[i].append(maxElement-matrix[i][j])
	return cost_matrix

	
def squarizeMatrix(matrix, filler):
	dimX = len(matrix)
	dimY = len(matrix[0])
	if dimX < dimY:
		for i in range(dimY - dimX):
			matrix.append([filler]*dimY)
	return matrix	
	
	
def strMatrix(matrix, name = None):
	dimX = len(matrix)
	dimY = len(matrix[0])
	string = ""
	if name != None:
		string += name + 2 * "\n"
	for i in range(dimX):
		for j in range(dimY):
			string += "%8.2f " % (matrix[i][j])
		string += "\n"
	return string
	
	
def debugOut(output):
	if DEBUG:
		print(output)
	return	

	
###############################################################################
## The next section contains all classes concerning the handling of Areas 
	
class Area: # abstract class

	def __init__(self, iNumRings):
		if iNumRings < 0:
			arg = "Size of Area has to be >= Zero!"
			raise ValueError(arg)
		self.iNumRings = iNumRings

	def getNumRings(self):
		return self.iNumRings

	def getPlots(self):
		return self.getPlotsFromRings(0, self.iNumRings)

	def getPlotsFromRing(self, iRing):
		return self.getPlotsFromRings(iRing, iRing)

	def getPlotsFromRings(self, iFirstRing, iLastRing): # abstract (semi)
		if iFirstRing < 0:
			arg = "First argument has to be >= Zero!"
			raise ValueError(arg)
		if iLastRing < 0:
			arg = "Second argument has to be >= Zero!"
			raise ValueError(arg)
		if iFirstRing > iLastRing:
			arg = "First argument has to be <= second argument!"
			raise ValueError(arg) 
		if iLastRing > self.iNumRings:
			arg = "Second argument has to be <= Size of Area!"
			raise ValueError(arg)

	def getRingFromPlot(self, iX, iY): # abstract
		raise "Error: not implemented. Has to be overwritten in subclasses"

	def isContainsPlot(self, iX, iY):
		iRing = self.getRingFromPlot(iX, iY)
		if iRing > self.iNumRings: return False
		return True
	

class SquareArea(Area):

	def getPlotsFromRings(self, iFirstRing, iLastRing):
		Area.getPlotsFromRings(self, iFirstRing, iLastRing)
		coords = []
		for x in range(-iLastRing, iLastRing+1):
			for y in range(-iLastRing, iLastRing+1):
				if not (abs(x) < iFirstRing and abs(y) < iFirstRing):
					coords.append((x,y))
		return coords

	def getRingFromPlot(self, iX, iY):
		return max(abs(iX), abs(iY))
	

class FatCrossArea(Area):

	def getPlotsFromRings(self, iFirstRing, iLastRing):
		Area.getPlotsFromRings(self, iFirstRing, iLastRing)
		coords = []		
		for x in range(-iLastRing, iLastRing+1):
			for y in range(-iLastRing, iLastRing+1):
				dist = abs(x) + abs(y)
				if dist > iFirstRing and dist < iLastRing+2:
					coords.append((x,y))
				elif dist == iFirstRing and (x == 0 or y == 0):
					coords.append((x,y))
		return coords

	def getRingFromPlot(self, iX, iY):
		dist = abs(iX) + abs(iY)
		if iX == 0 or iY == 0: return dist
		return dist - 1


class SearchArea:

	FULL_WEIGHT = 1.0

	def __init__(self, pHeartLand, pHinterLand):
		if not isinstance(pHeartLand, Area):
			arg = "pHeartLand has to be of type Area!"
			raise ValueError(arg)                    
		if not isinstance(pHinterLand, Area):
			arg = "pHinterLand has to be of type Area!"
			raise ValueError(arg) 
		if pHeartLand.getNumRings() > pHinterLand.getNumRings():
			arg = "Size of HeartLand > HinterLand!"
			raise ValueError(arg)
		self.pHeartLand = pHeartLand
		self.pHinterLand = pHinterLand

	def getPlots(self):
		lAllPlots = []
		iNumRings = self.pHinterLand.getNumRings()
		for x in range(-iNumRings, iNumRings + 1):
			for y in range(-iNumRings, iNumRings + 1):
				if self.pHeartLand.isContainsPlot(x,y) \
				or self.pHinterLand.isContainsPlot(x,y):
					lAllPlots.append((x,y))
		return lAllPlots		

	def getHeartLandPlots(self):
		return self.pHeartLand.getPlots()
		
	def getHinterLandBeltPlots(self):
		lBeltPlots = []
		iNumRings = self.pHinterLand.getNumRings()
		for x in range(-iNumRings, iNumRings + 1):
			for y in range(-iNumRings, iNumRings + 1):
				if self.pHinterLand.isContainsPlot(x,y) \
				and not self.pHeartLand.isContainsPlot(x,y):
					lBeltPlots.append((x,y))
		return lBeltPlots

	def getMaxHinterLandBeltSize(self):
		lBeltPlots = self.getHinterLandBeltPlots()
		iHinterLandSize = self.pHinterLand.getNumRings()
		iMaxBeltSize = 0
		for x, y in lBeltPlots:
			iHinterLandRing = self.pHinterLand.getRingFromPlot(x,y)
			iBeltSize = iHinterLandSize + 1 - iHinterLandRing
			if iBeltSize > iMaxBeltSize:
				iMaxBeltSize = iBeltSize
		return iMaxBeltSize

	def getPlotWeight(self, iX, iY):
		if self.pHeartLand.isContainsPlot(iX, iY):
			return self.FULL_WEIGHT
		if self.pHinterLand.isContainsPlot(iX, iY):
			iNumRings = self.pHinterLand.getNumRings()
			iRing = self.pHinterLand.getRingFromPlot(iX, iY)
			iMaxBeltSize = self.getMaxHinterLandBeltSize()
			fDistanceWeight = (iNumRings - iRing + 1) / float(iMaxBeltSize+1)
			return fDistanceWeight * self.FULL_WEIGHT
		return 0.0 # plot even outside hinterland
	
###############################################################################
## The following class provides Munkres' Optimization Algorithm for the 
## Linear Weighted Assignment Problem by Brian M. Clapper (c)

class Munkres:
    """
Calculate the Munkres solution to the classical assignment problem.
See the module documentation for usage.
    """

    def __init__(self):
        """Create a new instance"""
        self.C = None
        self.row_covered = []
        self.col_covered = []
        self.n = 0
        self.Z0_r = 0
        self.Z0_c = 0
        self.marked = None
        self.path = None

    def make_cost_matrix(profit_matrix, inversion_function):
        """
        Create a cost matrix from a profit matrix by calling
        'inversion_function' to invert each value. The inversion
        function must take one numeric argument (of any type) and return
        another numeric argument which is presumed to be the cost inverse
        of the original profit.

        This is a static method. Call it like this:

        cost_matrix = Munkres.make_cost_matrix(matrix, inversion_func)

        For example:

        cost_matrix = Munkres.make_cost_matrix(matrix, lambda x : sys.maxint - x)
        """
        cost_matrix = []
        for row in profit_matrix:
            cost_row = []
            for value in row:
                cost_row += [inversion_function(value)]
            cost_matrix += [cost_row]
        return cost_matrix

    make_cost_matrix = staticmethod(make_cost_matrix)

    def compute(self, cost_matrix):
        """
        Compute the indexes for the lowest-cost pairings between rows and
        columns in the database. Returns a list of (row, column) tuples
        that can be used to traverse the matrix.

        The matrix must be square.
        """
        self.C = self.__copy_matrix(cost_matrix)
        self.n = len(cost_matrix)
        self.row_covered = [False for i in range(self.n)]
        self.col_covered = [False for i in range(self.n)]
        self.Z0_r = 0
        self.Z0_c = 0
        self.path = self.__make_matrix(self.n * 2, 0)
        self.marked = self.__make_matrix(self.n, 0)

        done = False
        step = 1

        steps = { 1 : self.__step1,
                  2 : self.__step2,
                  3 : self.__step3,
                  4 : self.__step4,
                  5 : self.__step5,
                  6 : self.__step6 }

        while not done:
            try:
                func = steps[step]
                #print 'calling ' + str(func)
                step = func()
            except KeyError:
                done = True

        # Look for the starred columns
        results = []
        for i in range(self.n):
            for j in range(self.n):
                if self.marked[i][j] == 1:
                    results += [(i, j)]
        assert(len(results) == self.n)

        return results

    def __copy_matrix(self, matrix):
        """Return an exact copy of the supplied matrix"""
        copy = []
        for row in matrix:
            new_row = []
            for item in row:
                new_row += [item]
            copy += [new_row]
        return copy

    def __make_matrix(self, n, val):
        """Create an NxN matrix, populating it with the specific value."""
        matrix = []
        for i in range(n):
            matrix += [[val for j in range(n)]]
        return matrix

    def __step1(self):
        """
        For each row of the matrix, find the smallest element and
        subtract it from every element in its row. Go to Step 2.
        """
        C = self.C
        n = self.n
        for i in range(n):
            minval = self.C[i][0]
            # Find the minimum value for this row
            for j in range(n):
                if minval > self.C[i][j]:
                    minval = self.C[i][j]

            # Subtract that minimum from every element in the row.
            for j in range(n):
                self.C[i][j] -= minval

        return 2

    def __step2(self):
        """
        Find a zero (Z) in the resulting matrix. If there is no starred
        zero in its row or column, star Z. Repeat for each element in the
        matrix. Go to Step 3.
        """
        n = self.n
        for i in range(n):
            for j in range(n):
                if (self.C[i][j] == 0) and \
                   (not self.col_covered[j]) and \
                   (not self.row_covered[i]):
                    self.marked[i][j] = 1
                    self.col_covered[j] = True
                    self.row_covered[i] = True

        self.__clear_covers()
        return 3

    def __step3(self):
        """
        Cover each column containing a starred zero. If K columns are
        covered, the starred zeros describe a complete set of unique
        assignments. In this case, Go to DONE, otherwise, Go to Step 4.
        """
        n = self.n
        count = 0
        for i in range(n):
            for j in range(n):
                if self.marked[i][j] == 1:
                    self.col_covered[j] = True
                    count += 1

        if count >= n:
            step = 7 # done
        else:
            step = 4

        return step

    def __step4(self):
        """
        Find a noncovered zero and prime it. If there is no starred zero
        in the row containing this primed zero, Go to Step 5. Otherwise,
        cover this row and uncover the column containing the starred
        zero. Continue in this manner until there are no uncovered zeros
        left. Save the smallest uncovered value and Go to Step 6.
        """
        step = 0
        done = False
        row = -1
        col = -1
        star_col = -1
        while not done:
            (row, col) = self.__find_a_zero()
            if row < 0:
                done = True
                step = 6
            else:
                self.marked[row][col] = 2
                star_col = self.__find_star_in_row(row)
                if star_col >= 0:
                    col = star_col
                    self.row_covered[row] = True
                    self.col_covered[col] = False
                else:
                    done = True
                    self.Z0_r = row
                    self.Z0_c = col
                    step = 5

        return step

    def __step5(self):
        """
        Construct a series of alternating primed and starred zeros as
        follows. Let Z0 represent the uncovered primed zero found in Step 4.
        Let Z1 denote the starred zero in the column of Z0 (if any).
        Let Z2 denote the primed zero in the row of Z1 (there will always
        be one). Continue until the series terminates at a primed zero
        that has no starred zero in its column. Unstar each starred zero
        of the series, star each primed zero of the series, erase all
        primes and uncover every line in the matrix. Return to Step 3
        """
        count = 0
        path = self.path
        path[count][0] = self.Z0_r
        path[count][1] = self.Z0_c
        done = False
        while not done:
            row = self.__find_star_in_col(path[count][1])
            if row >= 0:
                count += 1
                path[count][0] = row
                path[count][1] = path[count-1][1]
            else:
                done = True

            if not done:
                col = self.__find_prime_in_row(path[count][0])
                count += 1
                path[count][0] = path[count-1][0]
                path[count][1] = col

        self.__convert_path(path, count)
        self.__clear_covers()
        self.__erase_primes()
        return 3

    def __step6(self):
        """
        Add the value found in Step 4 to every element of each covered
        row, and subtract it from every element of each uncovered column.
        Return to Step 4 without altering any stars, primes, or covered
        lines.
        """
        minval = self.__find_smallest()
        for i in range(self.n):
            for j in range(self.n):
                if self.row_covered[i]:
                    self.C[i][j] += minval
                if not self.col_covered[j]:
                    self.C[i][j] -= minval
        return 4

    def __find_smallest(self):
        """Find the smallest uncovered value in the matrix."""
        minval = sys.maxint
        for i in range(self.n):
            for j in range(self.n):
                if (not self.row_covered[i]) and (not self.col_covered[j]):
                    if minval > self.C[i][j]:
                        minval = self.C[i][j]
        return minval

    def __find_a_zero(self):
        """Find the first uncovered element with value 0"""
        row = -1
        col = -1
        i = 0
        n = self.n
        done = False

        while not done:
            j = 0
            while True:
                if (self.C[i][j] == 0) and \
                   (not self.row_covered[i]) and \
                   (not self.col_covered[j]):
                    row = i
                    col = j
                    done = True
                j += 1
                if j >= n:
                    break
            i += 1
            if i >= n:
                done = True

        return (row, col)

    def __find_star_in_row(self, row):
        """
        Find the first starred element in the specified row. Returns
        the column index, or -1 if no starred element was found.
        """
        col = -1
        for j in range(self.n):
            if self.marked[row][j] == 1:
                col = j
                break

        return col

    def __find_star_in_col(self, col):
        """
        Find the first starred element in the specified row. Returns
        the row index, or -1 if no starred element was found.
        """
        row = -1
        for i in range(self.n):
            if self.marked[i][col] == 1:
                row = i
                break

        return row

    def __find_prime_in_row(self, row):
        """
        Find the first prime element in the specified row. Returns
        the column index, or -1 if no starred element was found.
        """
        col = -1
        for j in range(self.n):
            if self.marked[row][j] == 2:
                col = j
                break

        return col

    def __convert_path(self, path, count):
        for i in range(count+1):
            if self.marked[path[i][0]][path[i][1]] == 1:
                self.marked[path[i][0]][path[i][1]] = 0
            else:
                self.marked[path[i][0]][path[i][1]] = 1

    def __clear_covers(self):
        """Clear all covered matrix cells"""
        for i in range(self.n):
            self.row_covered[i] = False
            self.col_covered[i] = False

    def __erase_primes(self):
        """Erase all prime markings"""
        for i in range(self.n):
            for j in range(self.n):
                if self.marked[i][j] == 2:
                    self.marked[i][j] = 0
	
'''	
CREDITS & COPYRIGHTS:
"FlavorMod" is inspired by dreiche2's "FfH_flavor_mapmod". Thanks Dreiche2, you
saved me a lot of work in working out the basics. The linear weighted assignemt 
algorithm (munkres), the heart of the optimization taking place here, is a 
Python module developed by Brian M. Clapper (bmc@clapper.org, 
http://www.clapper.org/bmc/). It is copyright (c) 2008 Brian M. Clapper.
'''














